/*****************************************************************************
 * gyach.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * Copyright (C) 2000 Chris Pinkham
 * Released under the terms of the GPL.
 * *NO WARRANTY*
 *
 * cpinkham@infi.net
 * http://www4.infi.net/~cpinkham/gyach/
 *****************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <regex.h>
#include <glib.h>

#include "parsecfg.h"
#include "gyach.h"
#include "images.h"
#include "yahoolib.h"
#include "commands.h"

#include "interface.h"
#include "support.h"
#include "nix.h"

/* this needs to be down here since it depends on gyach.h to define this */
#ifdef SHOW_STATUS_PIXMAPS
#include <gdk_imlib.h>
#endif

int debug_level = 00;
int debug_packets = 0;

GtkWidget * login_window;
GtkWidget * chat_window;
GtkWidget * chat_text;
GtkWidget * chat_entry;
GtkWidget * chat_emotions;
GtkWidget * chat_users;
GtkWidget * chat_user_menu;
GtkWidget * setup_menu;
GtkWidget * emote_menu;
GtkWidget * emote_window;
GtkWidget * pm_window;
GtkWidget * pm_entry;
GtkWidget * pm_user;
GtkWidget * chat_username;
GtkWidget * chat_password;
GtkWidget * chat_roomname;
GtkWidget * profile_window = NULL;
GtkWidget * about_box = NULL;
GtkWidget * setup_window = NULL;
GtkWidget * room_window = NULL;

GtkStyle  *profile_style = NULL;

int  emotion_count;
char *user_selected = NULL;
char *room_selected = NULL;
int  emote_selected = -1;
char *follow_user = NULL;
char *capture_filename = NULL;
FILE *capture_fp = NULL;

int child_pid = 0;
int my_status = 0;
time_t auto_away_time;
time_t child_timeout;

int  show_blended_colors = 0;
int  show_colors = 1;
int  show_fonts = 1;
int  def_user = -1;
int  auto_away = 15;
int  auto_login = 0;
int  pm_brings_back = 0;
int  buffer_size = 20;
int  show_html = 1;
int  show_statuses = 1;
int  show_enters = 1;
int  profile_viewer = 0;
int  ignore_on_mults = 0;
int  use_bold = 0;
int  use_underline = 0;
int  use_italics = 0;
int  font_size = -1;
int  force_lowercase = 0;
int  ping_minutes = 10;
char *browser_command = NULL;
char *font_name = NULL;
char *font_family = NULL;
char *use_color = NULL;
char *custom_away_message = NULL;
char *username0 = NULL;
char *username1 = NULL;
char *username2 = NULL;
char *username3 = NULL;
char *username4 = NULL;
char *password0 = NULL;
char *password1 = NULL;
char *password2 = NULL;
char *password3 = NULL;
char *password4 = NULL;
char *def_room  = NULL;
char *fav_room0 = NULL;
char *fav_room1 = NULL;
char *fav_room2 = NULL;
char *fav_room3 = NULL;
char *fav_room4 = NULL;
char *fav_room5 = NULL;
char *fav_room6 = NULL;
char *fav_room7 = NULL;
char *fav_room8 = NULL;
char *fav_room9 = NULL;

char cfg_filename[ 1024 ] = "";
char cfg_dir[1024] = "";

char username[41] = "";
char password[41] = "";
char roomname[129] = "";

int connected = 0;
int valid_user = 0;

GList *ignore_list = NULL;
GList *mute_list = NULL;
GList *regex_list = NULL;

int history_limit = 50;
int history_size = 0;
int history_current = -1;

GList *history_list = NULL;

SESSION *GlobalSessionHook;

void append_to_textbox(GtkWidget * window, GtkWidget * textbox, char *text);

cfgStruct cfg[] = {
	/* parameter  type address of variable */
	{ "show_blended_colors",	CFG_INT,		&show_blended_colors },
	{ "show_colors",			CFG_INT,		&show_colors },
	{ "show_fonts",				CFG_INT,		&show_fonts },
	{ "debug_packets",			CFG_INT,		&debug_packets },
	{ "def_user",				CFG_INT,		&def_user },
	{ "auto_away",				CFG_INT,		&auto_away },
	{ "auto_login",				CFG_INT,		&auto_login },
	{ "pm_brings_back",			CFG_INT,		&pm_brings_back },
	{ "buffer_size",			CFG_INT,		&buffer_size },
	{ "show_html",				CFG_INT,		&show_html },
	{ "show_statuses",			CFG_INT,		&show_statuses },
	{ "show_enters",			CFG_INT,		&show_enters },
	{ "profile_viewer",			CFG_INT,		&profile_viewer },
	{ "ignore_on_mults",		CFG_INT,		&ignore_on_mults },
	{ "use_bold",				CFG_INT,		&use_bold },
	{ "use_italics",			CFG_INT,		&use_italics },
	{ "use_underline",			CFG_INT,		&use_underline },
	{ "force_lowercase",		CFG_INT,		&force_lowercase },
	{ "ping_minutes",			CFG_INT,		&ping_minutes },
	{ "use_color",				CFG_STRING,		&use_color },
	{ "custom_away_message",	CFG_STRING,		&custom_away_message },
	{ "browser_command",		CFG_STRING,		&browser_command },
	{ "font_name",				CFG_STRING,		&font_name },
	{ "font_family",			CFG_STRING,		&font_family },
	{ "font_size",				CFG_INT,		&font_size },
	{ "username0",				CFG_STRING,		&username0 },
	{ "username1",				CFG_STRING,		&username1 },
	{ "username2",				CFG_STRING,		&username2 },
	{ "username3",				CFG_STRING,		&username3 },
	{ "username4",				CFG_STRING,		&username4 },
	{ "password0",				CFG_STRING,		&password0 },
	{ "password1",				CFG_STRING,		&password1 },
	{ "password2",				CFG_STRING,		&password2 },
	{ "password3",				CFG_STRING,		&password3 },
	{ "password4",				CFG_STRING,		&password4 },
	{ "def_room",				CFG_STRING,		&def_room },
	{ "fav_room0",				CFG_STRING,		&fav_room0 },
	{ "fav_room1",				CFG_STRING,		&fav_room1 },
	{ "fav_room2",				CFG_STRING,		&fav_room2 },
	{ "fav_room3",				CFG_STRING,		&fav_room3 },
	{ "fav_room4",				CFG_STRING,		&fav_room4 },
	{ "fav_room5",				CFG_STRING,		&fav_room5 },
	{ "fav_room6",				CFG_STRING,		&fav_room6 },
	{ "fav_room7",				CFG_STRING,		&fav_room7 },
	{ "fav_room8",				CFG_STRING,		&fav_room8 },
	{ "fav_room9",				CFG_STRING,		&fav_room9 },
	{ NULL,						CFG_END,		NULL }
};

/* fixme */
void no_home_dir_message() {
	/* print out a message or open a window or something */
}

gint gstrcmp( gpointer a, gpointer b ) {
	if (( ! a ) &&
		( ! b ))
		return( 0 );

	if ( ! a )
		return( -1 );

	if ( ! b )
		return( 1 );

	return( strcasecmp( a, b ));
}

void gyach_init() {
	read_config();
	ignore_load();

	history_list = g_list_append( history_list, "" );
	history_size = 1;

	auto_away_time = time(NULL) + ( auto_away * 60 );

	use_color = strdup( "black" );

	if ( ! font_name )
		font_name = strdup(
			"-*-times-medium-r-normal-*-*-*-*-*-*-*-*-*" );

	if ( ! font_family )
		font_family = strdup( "times" );

	if ( font_size < 0 )
		font_size = 14;

	if ( ! browser_command ) {
		browser_command =
			strdup( "netscape -noraise -remote openURL(%s,Gyach)" );
	}
}

void read_config() {
	struct stat sbuf;

	if ( ! getenv( "HOME" )) {
		no_home_dir_message();
		return;
	}

	if ( cfg_filename[ 0 ] == '\0' )
		sprintf( cfg_filename, "%s/.gyach/gyachrc", getenv( "HOME" ));

	if ( stat( cfg_filename, &sbuf ))
		write_config();

	cfgParse( cfg_filename, cfg, CFG_SIMPLE );
}

void write_config() {
	FILE *gyachrc;
	struct stat sbuf;

	if ( ! getenv( "HOME" )) {
		no_home_dir_message();
		return;
	}

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	if ( stat( cfg_dir, &sbuf ))
		mkdir( cfg_dir, 0700 );

	if ( cfg_filename[ 0 ] == '\0' )
		sprintf( cfg_filename, "%s/.gyach/gyachrc", getenv( "HOME" ));

	gyachrc = fopen( cfg_filename, "w" );

	if ( gyachrc ) {
		if (username0) fprintf( gyachrc, "username0 = \"%s\"\n", username0 );
		if (username1) fprintf( gyachrc, "username1 = \"%s\"\n", username1 );
		if (username2) fprintf( gyachrc, "username2 = \"%s\"\n", username2 );
		if (username3) fprintf( gyachrc, "username3 = \"%s\"\n", username3 );
		if (username4) fprintf( gyachrc, "username4 = \"%s\"\n", username4 );
		if (password0) fprintf( gyachrc, "password0 = \"%s\"\n", password0 );
		if (password1) fprintf( gyachrc, "password1 = \"%s\"\n", password1 );
		if (password2) fprintf( gyachrc, "password2 = \"%s\"\n", password2 );
		if (password3) fprintf( gyachrc, "password3 = \"%s\"\n", password3 );
		if (password4) fprintf( gyachrc, "password4 = \"%s\"\n", password4 );
		if (fav_room0) fprintf( gyachrc, "fav_room0 = \"%s\"\n", fav_room0 );
		if (fav_room1) fprintf( gyachrc, "fav_room1 = \"%s\"\n", fav_room1 );
		if (fav_room2) fprintf( gyachrc, "fav_room2 = \"%s\"\n", fav_room2 );
		if (fav_room3) fprintf( gyachrc, "fav_room3 = \"%s\"\n", fav_room3 );
		if (fav_room4) fprintf( gyachrc, "fav_room4 = \"%s\"\n", fav_room4 );
		if (fav_room5) fprintf( gyachrc, "fav_room5 = \"%s\"\n", fav_room5 );
		if (fav_room6) fprintf( gyachrc, "fav_room6 = \"%s\"\n", fav_room6 );
		if (fav_room7) fprintf( gyachrc, "fav_room7 = \"%s\"\n", fav_room7 );
		if (fav_room8) fprintf( gyachrc, "fav_room8 = \"%s\"\n", fav_room8 );
		if (fav_room9) fprintf( gyachrc, "fav_room9 = \"%s\"\n", fav_room9 );
		if (custom_away_message) fprintf( gyachrc,
			"custom_away_message = \"%s\"\n", custom_away_message );
		if (use_color) fprintf( gyachrc,
			"use_color = \"%s\"\n", use_color );
		if (font_name) fprintf( gyachrc,
			"font_name = \"%s\"\n", font_name );
		if (font_family) fprintf( gyachrc,
			"font_family = \"%s\"\n", font_family );
		if (browser_command) fprintf( gyachrc,
			"browser_command = \"%s\"\n", browser_command );

		fprintf( gyachrc, "show_blended_colors = %d\n", show_blended_colors );
		fprintf( gyachrc, "show_colors = %d\n", show_colors );
		fprintf( gyachrc, "show_fonts = %d\n", show_fonts );
		fprintf( gyachrc, "debug_packets = %d\n", debug_packets );
		fprintf( gyachrc, "auto_away = %d\n", auto_away );
		fprintf( gyachrc, "auto_login = %d\n", auto_login );
		fprintf( gyachrc, "pm_brings_back = %d\n", pm_brings_back );
		fprintf( gyachrc, "buffer_size = %d\n", buffer_size );
		fprintf( gyachrc, "show_html = %d\n", show_html );
		fprintf( gyachrc, "show_statuses = %d\n", show_statuses );
		fprintf( gyachrc, "show_enters = %d\n", show_enters );
		fprintf( gyachrc, "profile_viewer = %d\n", profile_viewer );
		fprintf( gyachrc, "ignore_on_mults = %d\n", ignore_on_mults );
		fprintf( gyachrc, "use_bold = %d\n", use_bold );
		fprintf( gyachrc, "use_italics = %d\n", use_italics );
		fprintf( gyachrc, "use_underline = %d\n", use_underline );
		fprintf( gyachrc, "force_lowercase = %d\n", force_lowercase );
		fprintf( gyachrc, "ping_minutes = %d\n", ping_minutes );
		fprintf( gyachrc, "font_size = %d\n", font_size );

		fclose( gyachrc );
	}
}

/*
 * emotion datafile fmt is "emotion]everyone|individual" similar to
 * CheetaChat but does NOT support %s for user's sex, only %N for other user's
 * name.
 */
void read_emotions() {
	gchar *listitem[3];
	FILE *fp;
	char line[1024];
	char *emotion = NULL;
	char *em_everyone = NULL;
	char *em_user = NULL;
	char *ptr = NULL;
	char emot_filename[1024] = "";
	int row;

	gtk_clist_set_sort_column(GTK_CLIST(chat_emotions), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(chat_emotions), 1 );

	gtk_clist_freeze(GTK_CLIST(chat_emotions));
	
	listitem[0] = emotion;
	listitem[1] = em_everyone;
	listitem[2] = em_user;
	emotion = line;

	if ( emot_filename[ 0 ] == '\0' )
		sprintf( emot_filename, "%s/.gyach/emotions.dat",
			getenv( "HOME" ));

	fp = fopen( emot_filename, "r" );

	emotion_count = 0;

	if ( fp ) {
		while( fgets( line, 1023, fp )) {
			if ( ! strchr( line, ']' ))	continue;

			while(( line[strlen(line)-1] == '\r' ) ||
				  ( line[strlen(line)-1] == '\n' )) {
				line[strlen(line)-1] = '\0';
			}

			emotion_count++;

			ptr = line;
			while( *ptr != ']' ) {
				*ptr = tolower( *ptr );
				ptr++;
			}

			line[0] = toupper( line[0] );
			ptr = strchr( line, ']' );
			*ptr = '\0';

			em_everyone = ptr + 1;
			ptr++;

			ptr = strchr( em_everyone, '|' );
			*ptr = '\0';
			em_user = ptr + 1;

			listitem[0] = emotion;
			row = gtk_clist_insert(GTK_CLIST(chat_emotions), 0, listitem );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 1, em_everyone );
			gtk_clist_set_text(GTK_CLIST(chat_emotions), row, 2, em_user );
		}
		fclose( fp );
	} else {
		/* no emotion file so set defaults */
	}

	gtk_clist_thaw(GTK_CLIST(chat_emotions));
}

void save_emotions() {
	gchar *em_name;
	gchar *em_everyone;
	gchar *em_user;
	char emot_filename[1024] = "";
	int i = 0;
	FILE *fp;

	if ( emot_filename[ 0 ] == '\0' )
		sprintf( emot_filename, "%s/.gyach/emotions.dat",
			getenv( "HOME" ));

	fp = fopen( emot_filename, "wb" );

	if ( fp ) {
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 0,&em_name)) {
				break;
			}

			gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 1, &em_everyone );
			gtk_clist_get_text(GTK_CLIST(chat_emotions), i, 2, &em_user );

			fprintf( fp, "%s]%s|%s\n", em_name, em_everyone, em_user );
			
			i++;
		}
		fclose( fp );
	}
}

void strip_html_tags( char *str ) {
	char *to, *from;

	to = str;
	from = str;

	while( *from ) {
		if ( *from == '<' ) {
			from++;
			while(( *from != '>' ) &&
				  ( *from != '<' )) {
				from++;
			}
			from++;
		} else {
			*to = *from;
			to++;
			from++;
		}
	}
	*to = *from;

	to = str;
	from = str;
	while( *from ) {
		if ( *from == '\r' ) {
			from++;
		} else if ( *from == '&' ) {
			if ( *(from+1) == '#' ) {
				while( *from != ';' ) {
					from++;
				}
			} else if (( *(from+1) == 'n' ) &&
					   ( *(from+2) == 'b' ) &&
					   ( *(from+3) == 's' ) &&
					   ( *(from+4) == 'p' ) &&
					   ( *(from+5) == ';' )) {
				from += 5;
				*to = ' ';
				to++;
			}
			from++;
		} else {
			*to = *from;
			to++;
			from++;
		}
	}
	*to = *from;

	to = str;
	from = str;
	while( *from ) {
		if (( *from == '\n' ) &&
			( *(from+1) == '\n' )) {
		} else {
			*to = *from;
			to++;
		}
		from++;
	}
	*to = *from;
}

char *download_image( char *image_url ) {
	static char tmp_filename[1025] = "/tmp/gyach.img";
	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	char result[65535];
	char tmp[65535];
	char *ptr;
	int nr = 0, tr = 0;
	int ifp;

	char host[75] = "";

	if ( strncmp( image_url, "http://", 7 ))
		return "";

	strncpy( host, image_url + 7, 74 );
	host[74] = '\0';
	ptr = strchr( host, '/' );
	*ptr = '\0';

	ptr = strchr( image_url + 7, '/' );

	hinfo = gethostbyname( host );

	if ( ! hinfo || ( hinfo->h_addrtype != AF_INET )) {
		/* fixme, show an error box */
		return "";
	}

	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	if (( sock = socket( AF_INET, SOCK_STREAM, 6 )) == -1 ) {
		/* fixme, error message */
		if ( debug_level > 10 ) {
			printf( "error in socket()\n" );
		}
		return "";
	} else {
		if ( connect( sock, (struct sockaddr*)&sa, sizeof(sa)) == -1 ) {
			/* fixme, error message */
			if ( debug_level > 10 ) {
				printf( "error in connect()\n" );
			}
			close(sock);
			return "";
		}
	}

	snprintf( url, 1024,
		"GET %s HTTP/1.0\r\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: gyach/?.?\r\n"
		"\r\n"
		"\r\n", ptr );

	write( sock, url, strlen( url ));

	result[0] = '\0';

	tr = 0;
	nr = read( sock, tmp, 512 );
	while(( nr > 0 ) &&
		( tr < 65535 )) {
		tmp[nr] = '\0';
		memcpy( result + tr, tmp, nr );
		tr += nr;

		while( gtk_events_pending())
			gtk_main_iteration();

		nr = read( sock, tmp, 512 );
	}

	close( sock );

	ptr = strstr( result, "\r\n\r\n" );
	ptr += 4;

	if ( strstr( result, "Not Found" ))
		return( "" );

	ifp = open( tmp_filename, O_CREAT | O_TRUNC| O_WRONLY, 0600 );
	write( ifp, ptr, tr - ( ptr - result ));
	close( ifp );

	return tmp_filename;
}

void display_url( char *url ) {
	char cmd[1025];

	if ( strlen( url ) == 0 )
		return;

	/* quick sanity check, fixme */
	if (( strchr( url, ';' )) ||
		( strchr( url, '`' ))) {
		return;
	}

	snprintf( cmd, 1024, browser_command, url );

	if( ! fork()) {
		system( cmd );
		_exit(0);
	}
}

void display_profile( char *user ) {
#ifdef SHOW_STATUS_PIXMAPS
	GdkImlibImage *orig_image;
	GdkImlibImage *scaled_image;
#endif

	int sock;
	struct sockaddr_in sa;
	struct hostent *hinfo;
	char url[1025];
	char result[32769];
	char tmp[32769];
	char buf[8192];
	int nr = 0, tr = 0;

	char *ptr;
	char pic_url[129] = "";
	char last_updated[21] = "";
	char email[129] = "";
	char realname[51] = "";
	char location[51] = "";
	char age[5] = "";
	char married[21] = "";
	char sex[11] = "";
	char occupation[51] = "";
	char online_now[11] = "";
	char homepage[129] = "";
	char coollink1[129] = "";
	char coollink2[129] = "";
	char coollink3[129] = "";

	GtkWidget	*tmp_widget;

	if ( ! profile_viewer ) {
		snprintf( homepage, 128, "http://profiles.yahoo.com/%s", user );
		display_url( homepage );
		return;
	}

	hinfo = gethostbyname( "profiles.yahoo.com" );

	if ( ! hinfo || ( hinfo->h_addrtype != AF_INET )) {
		/* fixme, show an error box */
		return;
	}

	memset( &sa, 0, sizeof(sa));
	memmove((void*)&sa.sin_addr.s_addr, hinfo->h_addr, hinfo->h_length );

	sa.sin_family = AF_INET;
	sa.sin_port = htons( 80 );

	if (( sock = socket( AF_INET, SOCK_STREAM, 6 )) == -1 ) {
		/* fixme, error message */
		if ( debug_level > 10 ) {
			printf( "error in socket()\n" );
		}
		return;
	} else {
		if ( connect( sock, (struct sockaddr*)&sa, sizeof(sa)) == -1 ) {
			/* fixme, error message */
			if ( debug_level > 10 ) {
				printf( "error in connect()\n" );
			}
			close(sock);
			return; 
		}
	}

	snprintf( url, 1024,
		"GET /%s HTTP/1.0\r\n"
		"Accept: text/html\r\n"
		"Accept: text/plain\r\n"
		"User-Agent: gyach/?.?\r\n"
		"\r\n"
		"\r\n", user );

	write( sock, url, strlen( url ));

	result[0] = '\0';

	tr = 0;
	nr = read( sock, tmp, 32768 );
	while( nr > 0 ) {
		tmp[nr] = '\0';
		strcat( result, tmp );
		tr += nr;
		nr = read( sock, tmp, 32768 - tr );
	}

	/* find the url to their pic */
	sprintf( tmp, "alt=\"%s\"", user );
	ptr = tmp;
	while( *ptr )
		*(ptr++) = tolower( *ptr );
	ptr = strstr( result, tmp );
	while(( *(ptr-5) != 's' ) || ( *(ptr-4) != 'r' ) || ( *(ptr-3) != 'c' ) ||
		  ( *(ptr-2) != '=' ) || ( *(ptr-1) != '"' )) {
		ptr--;
	}
	/* now we are at the beginning of the url, so strncpy some */
	strncpy( pic_url, ptr, 128 );
	ptr = strchr( pic_url, '"' );
	*ptr = '\0';

	/* strip html to make the rest easier */
	strip_html_tags( result );

	/* attempt to find some strings */
	ptr = strstr( result, "Last Updated:" );
	if ( ptr ) {
		ptr = strchr( ptr, ':' );
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( last_updated, ptr, 20 );
		ptr = strchr( last_updated, '\n' );
		*ptr = '\0';
	}

	ptr = strstr( result, "My Email" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( email, ptr, 128 );
		ptr = strchr( email, '\n' );
		*ptr = '\0';
	}

	ptr = strstr( result, "Real Name:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( realname, ptr, 50 );
		ptr = strchr( realname, '\n' );
		*ptr = '\0';

		if ( strchr( realname, ':' ))
			realname[0] = '\0';
	}

	ptr = strstr( result, "Location:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( location, ptr, 50 );
		ptr = strchr( location, '\n' );
		*ptr = '\0';

		if ( strchr( location, ':' ))
			location[0] = '\0';
	}

	ptr = strstr( result, "Age:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( age, ptr, 4 );
		ptr = strchr( age, '\n' );
		*ptr = '\0';

		if ( strchr( age, ':' ))
			age[0] = '\0';
	}

	ptr = strstr( result, "Marital Status:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( married, ptr, 20 );
		ptr = strchr( married, '\n' );
		*ptr = '\0';

		if ( strchr( married, ':' ))
			married[0] = '\0';
	}

	ptr = strstr( result, "Sex:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( sex, ptr, 10 );
		ptr = strchr( sex, '\n' );
		*ptr = '\0';

		if ( strchr( sex, ':' ))
			sex[0] = '\0';
	}

	ptr = strstr( result, "Occupation:" );
	if ( ptr ) {
		while( *ptr != '\n' )
			ptr++;
		while( *ptr == '\n' ) 
			ptr++;
		strncpy( occupation, ptr, 50 );
		ptr = strchr( occupation, '\n' );
		*ptr = '\0';

		if ( strchr( occupation, ':' ))
			occupation[0] = '\0';
	}

	ptr = strstr( result, "Home Page:" );
	if ( ptr ) {
		ptr = strchr( ptr, ':' );
		ptr++;
		ptr = skip_whitespace( ptr );
		strncpy( homepage, ptr, 128 );
		ptr = strchr( homepage, ' ' );
		*ptr = '\0';
	}

	ptr = strstr( result, "I'm online now!" );
	if ( ptr ) {
		strcpy( online_now, "Online now!" );
	} else {
		strcpy( online_now, "OFFline" );
	}

	if ( debug_level >= 10 ) {
		printf( "profile info for %s:\n", user );
		printf( "Picture         : '%s'\n", pic_url );
		printf( "Last Updated    : '%s'\n", last_updated );
		printf( "Email           : '%s'\n", email );
		printf( "Real Name       : '%s'\n", realname );
		printf( "Location        : '%s'\n", location );
		printf( "Age             : '%s'\n", age );
		printf( "Married?        : '%s'\n", married );
		printf( "Sex             : '%s'\n", sex );
		printf( "Occupation      : '%s'\n", occupation );
		printf( "Online Now      : '%s'\n", online_now );
		printf( "Home Page       : '%s'\n", homepage );
		printf( "Cool Link #1    : '%s'\n", coollink1 );
		printf( "Cool Link #2    : '%s'\n", coollink2 );
		printf( "Cool Link #3    : '%s'\n", coollink3 );
	}

	close(sock);


	if ( ! profile_window ) {
		profile_window = create_profile_window();
	}

	if ( ! profile_style ) {
		profile_style = gtk_style_new();
		profile_style->font = gdk_font_load( "fixed" );
	}

	tmp_widget = lookup_widget( profile_window, "profile_info" );
	gtk_widget_set_style( tmp_widget, profile_style );

	sprintf( buf,
		" Yahoo! ID      : %s\n"
		" Real Name      : %s\n"
		" Last Updated   : %s\n"
		" Location       : %s\n"
		" Age            : %s\n"
		" Marital Status : %s\n"
		" Sex            : %s\n"
		" Occupation     : %s\n"
		" Email          : %s\n"
		" HomePage       : %s\n"
		" Cool Links     : %s\n"
		"                  %s\n"
		"                  %s\n",
		user, realname, last_updated, location, age, married, sex, occupation,
		email, homepage, coollink1, coollink2, coollink3 );
	gtk_label_set_text( GTK_LABEL(tmp_widget), buf );

	gtk_widget_show( profile_window );

	tmp_widget = lookup_widget( profile_window, "user_picture" );

	ptr = download_image( pic_url );

/* for testing purposes
	system( "xv /tmp/gyach.img" );
*/

#ifdef SHOW_STATUS_PIXMAPS
	orig_image = gdk_imlib_load_image( ptr );

	scaled_image = gdk_imlib_clone_scaled_image( orig_image, 150, 150 );
	gdk_imlib_destroy_image( orig_image );
	unlink( ptr );

	gdk_imlib_save_image_to_ppm( scaled_image, ptr );
	gdk_imlib_destroy_image( scaled_image );

	gdk_imlib_load_file_to_pixmap( ptr, &(GTK_PIXMAP(tmp_widget)->pixmap),
		&(GTK_PIXMAP(tmp_widget)->mask) );
#endif

	unlink( ptr );
}

void lower_str( char *str ) {
	char *ptr = str;

	while( *ptr ) {
		*(ptr++) = tolower( *ptr );
	}
}

void edit_emote() {
	gchar *listitem[3];
	gchar emotion[128] = "";
	gchar em_everyone[128] = "";
	gchar em_user[128] = "";
	GtkWidget *emote_name;
	GtkWidget *emote_everyone;
	GtkWidget *emote_user;

	listitem[0] = emotion;
	listitem[1] = em_everyone;
	listitem[2] = em_user;

	if ( emote_selected > -1 ) {
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 0,
			listitem );
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 1,
			listitem+1 );
		gtk_clist_get_text( GTK_CLIST(chat_emotions), emote_selected, 2,
			listitem+2 );
	}

    if ( ! emote_window ) {
		emote_window = create_emote_window();
		gtk_widget_show( emote_window );
	}

	emote_name = lookup_widget( emote_window, "emote_name" );
	emote_everyone = lookup_widget( emote_window, "emote_everyone" );
	emote_user = lookup_widget( emote_window, "emote_user" );

	gtk_entry_set_text( GTK_ENTRY(emote_name), listitem[0] );
	gtk_entry_set_text( GTK_ENTRY(emote_everyone), listitem[1] );
	gtk_entry_set_text( GTK_ENTRY(emote_user), listitem[2] );
}

int find_user_row( char *user ) {
	char tmp_user[129];
	char tmp_list[129];
	char *listitem[1];
	int i;
	int r;

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	r = -1;
	for( i = 0 ; i < GTK_CLIST(chat_users)->rows && r == -1; i++ ) {
		gtk_clist_get_text( GTK_CLIST(chat_users), i, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );
		if ( ! strcmp( tmp_list, tmp_user ))
			r = i;
	}

	return( r );
}

int chars_match( char *str1, char *str2 ) {
	int i = 0;

	/* if the 2nd string is longer then it's can't be the same word */
	if ( strlen( str2 ) > strlen( str1 ) )
		return( 0 );

	while(( *str1 ) && ( *str2 ) && ( *str1 == *str2 )) {
		i++;
		str1++;
		str2++;
	}

	return( i );
}

/*
 * fixme, this function needs cleaning up probably
 */
char *find_user_best_match( char *part ) {
	static char best_match[129] = "";
	static char last_part[129] = "";
	static int  last_match = 0;
	char tmp_user[129];
	char tmp_list[129];
	char *listitem[1];
	int bm = last_match;


	if ( ! part[0] ) {
		best_match[0] = '\0';
		last_part[0] = '\0';
		return( best_match );
	}

	if (( ! last_part[0] ) || strncmp( part, last_part, strlen( last_part ))) {
		best_match[0] = '\0';
		last_match = 0;
		strcpy( last_part, part );
	} else {
		strcpy( part, last_part );
		last_match++;
	}

	if ( last_match >= GTK_CLIST(chat_users)->rows )
		last_match = 0;

	strcpy( tmp_user, part );
	lower_str( tmp_user );

	gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
	strcpy( tmp_list, listitem[0] );
	lower_str( tmp_list );

	while(( strncmp( tmp_list, part, strlen( part ))) &&
		  ( last_match < GTK_CLIST(chat_users)->rows )) {
		last_match++;
		gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );
	}

	/* did we hit the end of the list after starting at the beginning */
	/* if so, return */
	if (( last_match >= GTK_CLIST(chat_users)->rows ) &&
		( bm == 0 )) {
		strcpy( best_match, part );
		last_match = 0;
		return( best_match );
	} 

	/* if we don't match, try to start from the beginning again */
	if ( strncmp( tmp_list, part, strlen( part ))) {
		last_match = 0;
		gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1, listitem );
		strcpy( tmp_list, listitem[0] );
		lower_str( tmp_list );

		while(( strncmp( tmp_list, part, strlen( part ))) &&
			  ( last_match < bm )) {
			last_match++;
			gtk_clist_get_text( GTK_CLIST(chat_users), last_match, 1,
				listitem );
			strcpy( tmp_list, listitem[0] );
			lower_str( tmp_list );
		}
	}

	if ( strncmp( tmp_list, part, strlen( part ))) {
		strcpy( best_match, part );
		last_match = 0;
	} else {
		strcpy( best_match, tmp_list );
	}
		
	return( best_match );
}

void ignore_add( char *user ) {
	char tmp_user[129];

	if ( ! user )
		return;

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	if ( g_list_find_custom( ignore_list, tmp_user, (GCompareFunc)gstrcmp ))
		return;

	ignore_list = g_list_append( ignore_list, strdup( tmp_user ));
	g_list_sort( ignore_list, (GCompareFunc)gstrcmp );
}

void ignore_remove( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user )
		return;

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( ignore_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		ptr = this_user->data;
		ignore_list = g_list_remove( ignore_list, ptr );
		free( ptr );
	}
}

int ignore_check( char *user ) {
	GList *this_user;
	char tmp_user[129];

	if ( ! user )
		return( 0 );

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( ignore_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user )
		return( 1 );
	else
		return( 0 );
}

void regex_add( char *regex ) {
	char tmp_regex[129];

	if ( ! regex )
		return;

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	if ( g_list_find_custom( regex_list, tmp_regex, (GCompareFunc)gstrcmp ))
		return;

	regex_list = g_list_append( regex_list, strdup( tmp_regex ));
	g_list_sort( regex_list, (GCompareFunc)gstrcmp );
}

void regex_remove( char *regex ) {
	GList *this_regex;
	char tmp_regex[129];
	char *ptr;

	if ( ! regex )
		return;

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	this_regex = g_list_find_custom( regex_list, tmp_regex,
		(GCompareFunc)gstrcmp );

	if ( this_regex ) {
		ptr = this_regex->data;
		regex_list = g_list_remove( regex_list, ptr );
		free( ptr );
	}
}

int regex_check( char *regex ) {
	GList *this_regex;
	char tmp_regex[129];

	if ( ! regex )
		return( 0 );

	strcpy( tmp_regex, regex );
	lower_str( tmp_regex );

	this_regex = g_list_find_custom( regex_list, tmp_regex,
		(GCompareFunc)gstrcmp );

	if ( this_regex )
		return( 1 );
	else
		return( 0 );
}

int auto_ignore_check( char *str ) {
	GList *this_regex;
	regex_t ai_regex;
	

	if ( ! str )
		return( 0 );

	this_regex = regex_list;
	while( this_regex ) {
		regcomp( &ai_regex, this_regex->data, REG_ICASE | REG_NOSUB );
		if ( ! regexec( &ai_regex, str, 0, NULL, 0 )) {
			return( 1 );
		}

		this_regex = g_list_next( this_regex );
	}

	return( 0 );
}

void ignore_load() {
	GList *this_link;
	char if_filename[1024];
	char buf[129];
	FILE *fp;

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	sprintf( if_filename, "%s/ignore.list", cfg_dir );

	fp = fopen( if_filename, "r" );

	if ( fp ) {
		/* clear the list if there already is one */
		this_link = ignore_list;
		while( this_link ) {
			free( this_link->data );
			this_link = g_list_next( this_link );
		}

		g_list_free( ignore_list );

		ignore_list = NULL;

		/* load the new list */
		while( fgets( buf, 128, fp )) {
			/* chop the newline */
			buf[strlen(buf)-1] = '\0';

			ignore_list = g_list_append( ignore_list, strdup( buf ));
		}

		fclose( fp );

		g_list_sort( ignore_list, (GCompareFunc)gstrcmp );
	}

	sprintf( if_filename, "%s/ignore.regex", cfg_dir );

	fp = fopen( if_filename, "r" );

	if ( fp ) {
		/* clear the list if there already is one */
		this_link = regex_list;
		while( this_link ) {
			free( this_link->data );
			this_link = g_list_next( this_link );
		}

		g_list_free( regex_list );

		regex_list = NULL;

		/* load the new list */
		while( fgets( buf, 128, fp )) {
			/* chop the newline */
			buf[strlen(buf)-1] = '\0';

			regex_list = g_list_append( regex_list, strdup( buf ));
		}

		fclose( fp );

		g_list_sort( regex_list, (GCompareFunc)gstrcmp );
	}
}

void ignore_save() {
	GList *this_user;
	GList *this_regex;
	char if_filename[1024];
	FILE *fp;

	if ( cfg_dir[ 0 ] == '\0' )
		sprintf( cfg_dir, "%s/.gyach", getenv( "HOME" ));

	sprintf( if_filename, "%s/ignore.list", cfg_dir );

	fp = fopen( if_filename, "w" );

	if ( fp ) {
		this_user = ignore_list;
		while( this_user ) {
			fprintf( fp, "%s\n", (char *)this_user->data );

			this_user = g_list_next( this_user );
		}

		fclose( fp );
	}

	sprintf( if_filename, "%s/ignore.regex", cfg_dir );

	fp = fopen( if_filename, "w" );

	if ( fp ) {
		this_regex = regex_list;
		while( this_regex ) {
			fprintf( fp, "%s\n", (char *)this_regex->data );
			this_regex = g_list_next( this_regex );
		}

		fclose( fp );
	}
}

void ignore_toggle( char *user ) {
	GList *this_user;
	int r;
	int i;
	char buf[129] = "";
#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap;
	GdkBitmap *st_mask;
#else
	char status[5] = "";
#endif

	r = find_user_row( user );

	if ( user[0] ) {
		if ( ignore_check( user )) {
			ignore_remove( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				gdk_imlib_data_to_pixmap( pixmap_status_here, &st_pixmap,
					&st_mask );
#else
				strcpy( status, "" );
#endif
			}
			sprintf( buf, "*** no longer ignoring %s%s%s ***\n",
				GYACH_COLOR_BLUE, user, GYACH_COLOR_BLACK );
		} else {
			ignore_add( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				gdk_imlib_data_to_pixmap( pixmap_status_ignore, &st_pixmap,
					&st_mask );
#else
				strcpy( status, "IG" );
#endif
			}
			sprintf( buf, "*** ignoring %s%s%s ***\n",
				GYACH_COLOR_BLUE, user, GYACH_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	} else {
		append_to_textbox( chat_window, chat_text,
			"*** Ignoring the following users:\n" );
		i = 0;
		this_user = ignore_list;
		while( this_user ) {
			sprintf( buf, "\t%s\n", (char *)this_user->data );
			append_to_textbox( chat_window, chat_text, buf );
			i++;
			this_user = g_list_next( this_user );
		}
		if ( i ) {
			sprintf( buf, "*** Ignoring %d users ***\n", i );
			append_to_textbox( chat_window, chat_text, buf );
		} else {
			append_to_textbox( chat_window, chat_text, "\tNo Users Ignored\n" );
		}
	}

	ignore_save();

	if ( r < 0 )
		return;

#ifdef SHOW_STATUS_PIXMAPS
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), r, 0, st_pixmap, st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), r, 0, status );
#endif
}

void mute_add( char *user ) {
	char tmp_user[129];

	if ( ! user )
		return;

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	if ( g_list_find_custom( mute_list, tmp_user, (GCompareFunc)gstrcmp ))
		return;

	mute_list = g_list_append( mute_list, strdup( tmp_user ));
	g_list_sort( mute_list, (GCompareFunc)gstrcmp );
}

void mute_remove( char *user ) {
	GList *this_user;
	char tmp_user[129];
	char *ptr;

	if ( ! user )
		return;

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( mute_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user ) {
		ptr = this_user->data;
		mute_list = g_list_remove( mute_list, ptr );
		free( ptr );
	}
}

int mute_check( char *user ) {
	GList *this_user;
	char tmp_user[129];

	if ( ! user )
		return( 0 );

	strcpy( tmp_user, user );
	lower_str( tmp_user );

	this_user = g_list_find_custom( mute_list, tmp_user,
		(GCompareFunc)gstrcmp );

	if ( this_user )
		return( 1 );
	else
		return( 0 );
}

void mute_toggle( char *user ) {
	GList *this_user;
	int r;
	int i;
	char buf[129] = "";
#ifdef SHOW_STATUS_PIXMAPS
	GdkPixmap *st_pixmap;
	GdkBitmap *st_mask;
#else
	char status[5] = "";
#endif

	r = find_user_row( user );

	if ( user[0] ) {
		if ( mute_check( user )) {
			mute_remove( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				gdk_imlib_data_to_pixmap( pixmap_status_here, &st_pixmap,
					&st_mask );
#else
				strcpy( status, "" );
#endif
			}
			sprintf( buf, "*** no longer muting %s%s%s ***\n",
				GYACH_COLOR_BLUE, user, GYACH_COLOR_BLACK );
		} else {
			mute_add( user );
			if ( r >= 0 ) {
#ifdef SHOW_STATUS_PIXMAPS
				gdk_imlib_data_to_pixmap( pixmap_status_mute, &st_pixmap,
					&st_mask );
#else
				strcpy( status, "MU" );
#endif
			}
			sprintf( buf, "*** muting %s%s%s ***\n",
				GYACH_COLOR_BLUE, user, GYACH_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	} else {
		append_to_textbox( chat_window, chat_text,
			"*** Muting the following users:\n" );
		i = 0;
		this_user = mute_list;
		while( this_user ) {
			sprintf( buf, "\t%s\n", (char *)this_user->data );
			append_to_textbox( chat_window, chat_text, buf );
			i++;
			this_user = g_list_next( this_user );
		}
		if ( i ) {
			sprintf( buf, "*** Muting %d users ***\n", i );
			append_to_textbox( chat_window, chat_text, buf );
		} else {
			append_to_textbox( chat_window, chat_text, "\tNo Users Muted\n" );
		}
	}

	if ( r < 0 )
		return;

#ifdef SHOW_STATUS_PIXMAPS
	gtk_clist_set_pixmap(GTK_CLIST(chat_users), r, 0, st_pixmap, st_mask );
#else
	gtk_clist_set_text(GTK_CLIST(chat_users), r, 0, status );
#endif
}

void regex_toggle( char *regex ) {
	char buf[129] = "";

	if ( regex[0] ) {
		if ( regex_check( regex )) {
			regex_remove( regex );
			sprintf( buf, "*** no longer auto-ignoring on regex '%s%s%s' ***\n",
				GYACH_COLOR_BLUE, regex, GYACH_COLOR_BLACK );
		} else {
			regex_add( regex );
			sprintf( buf, "*** auto-ignoring on regex '%s%s%s' ***\n",
				GYACH_COLOR_BLUE, regex, GYACH_COLOR_BLACK );
		}
		append_to_textbox( chat_window, chat_text, buf );
	}

	ignore_save();
}

/*
 * dump a yin packet out to stdout for testing purposes
 */
void dump_packet() {
	int i = 0;
	int x = 0;
	int y = 0;
	char *ptr;

	printf( "type = 0x%x, size = %d\n",
		GlobalSessionHook->yin.type,
		GlobalSessionHook->yin.size );

	ptr = GlobalSessionHook->yin.data;
	while( i < GlobalSessionHook->yin.size ) {
		if ( x < 16 ) {
			printf( "%02x ", *(ptr+i) & 0xFF );
			x++;
			i++;
		} else {
			printf( "   " );
			for( ; x > 0 ; x-- ) {
				if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
					printf( "%c", *(ptr+i-x) );
				} else {
					printf( "." );
				}
			}

			printf( "\n" );
			x = 0;
		}
	}
	for( y = x; y < 16 ; y++ ) {
		printf( "   " );
	}
	printf( "   " );
	for( ; x > 0 ; x-- ) {
		if ( isgraph( *(ptr+i-x) ) || *(ptr+i-x)  == ' ' ) {
			printf( "%c", *(ptr+i-x) );
		} else {
			printf( "." );
		}
	}

	printf( "\n" );
}

/*
 * attempt to highlight urls if they aren't already highlighted along with
 * turning html style tags into esc sequences.
 */
void convert_tags( char *str ) {
	char tmp[4097];
	char *from;
	char *to;
	char *ptr;
	int len = 0;

	/* return if too long */
	if ( strlen( str ) > 4096 )
		return;

	/* convert html tags to ESC sequences */
	if ( strchr( str, '<' )) {
		memset( tmp, 0, sizeof( tmp ));

		from = str;
		to = tmp;

		ptr = strchr( from, '<' );
		while( ptr ) {
			len = ptr - from;
			strncpy( to, from, len );
			to += len;
			*to = '\0';

			from += len;

			if ( !strncasecmp( from, "<red>", 5 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_RED );
					to += strlen( GYACH_COLOR_RED );
				}
				from += 5;
			} else if ( !strncasecmp( from, "<blue>", 6 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_BLUE );
					to += strlen( GYACH_COLOR_BLUE );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<black>", 7 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_BLACK );
					to += strlen( GYACH_COLOR_BLACK );
				}
				from += 7;
			} else if ( !strncasecmp( from, "<gray>", 6 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_YELLOW );
					to += strlen( GYACH_COLOR_YELLOW );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<green>", 7 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_GREEN );
					to += strlen( GYACH_COLOR_GREEN );
				}
				from += 7;
			} else if ( !strncasecmp( from, "<pink>", 6 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_CYAN );
					to += strlen( GYACH_COLOR_CYAN );
				}
				from += 6;
			} else if ( !strncasecmp( from, "<purple>", 8 )) {
				if ( show_html ) {
					strcat( to, GYACH_COLOR_MAGENTA );
					to += strlen( GYACH_COLOR_MAGENTA );
				}
				from += 8;
			} else if ( !strncasecmp( from, "<i>", 3 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_ITALICON );
					to += strlen( YAHOO_STYLE_ITALICON );
				}
				from += 3;
			} else if ( !strncasecmp( from, "</i>", 4 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_ITALICOFF );
					to += strlen( YAHOO_STYLE_ITALICOFF );
				}
				from += 4;
			} else if ( !strncasecmp( from, "<b>", 3 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_BOLDON );
					to += strlen( YAHOO_STYLE_BOLDON );
				}
				from += 3;
			} else if ( !strncasecmp( from, "</b>", 4 )) {
				if ( show_html ) {
					strcat( to, YAHOO_STYLE_BOLDOFF );
					to += strlen( YAHOO_STYLE_BOLDOFF );
				}
				from += 4;
			} else if ( !strncasecmp( from, "</font>", 7 )) {
				/* skip ending font tags */
				from += 7;
			} else if ( !strncasecmp( from, "<alt", 4 )) {
				/* skip alt tags */
				from += 4;
				while(( *from ) && ( *from != '>' ))
					from++;
				from++;
			} else if ( !strncasecmp( from, "</alt>", 6 )) {
				/* skip ending alt tags */
				from += 6;
			} else if ( !strncasecmp( from, "<url=", 5 )) {
				/* skip beginning url tags */
				ptr = from + 5;
				while(( *ptr ) && ( *ptr != '>' )) {
					*(to++) = *(ptr)++;
				}
				if ( *ptr )
					from = ptr + 1;
				*(to++) = ' ';
				*(to++) = '(';
			} else if ( !strncasecmp( from, "</url>", 6 )) {
				/* skip ending url tags */
				*(to++) = ')';
				from += 6;
			} else if ( !strncasecmp( from, "<snd", 4 )) {
				/* skip beginning snd tags */
				ptr = from + 4;
				while(( *ptr ) && ( *ptr != '>' ))
					ptr++;
				if ( *ptr )
					from = ptr + 1;
			} else if ( !strncasecmp( from, "</snd>", 6 )) {
				/* skip ending snd tags */
				from += 6;
			} else if ( !strncasecmp( from, "<fade", 5 )) {
				/* skip beginning fade tags */
				ptr = from + 5;
				while(( *ptr ) && ( *ptr != '>' ))
					ptr++;
				if ( *ptr )
					from = ptr + 1;
			} else if ( !strncasecmp( from, "</fade>", 7 )) {
				/* skip ending fade tags */
				from += 7;
			} else {
				*(to++) = *(from++);
			}

			ptr = strchr( from, '<' );
		}

		strcat( to, from );
		to += strlen( from );
		*to = '\0';
		strcpy( str, tmp );
	}

	/* return if no urls */
	if ( ! strstr( str, "http://" ))
		return;

	/* return if already marked */
	if ( strstr( str, YAHOO_STYLE_URLON ))
		return;

	memset( tmp, 0, sizeof( tmp ));

	from = str;
	to = tmp;

	ptr = strstr( str, "http://" );

	while( ptr ) {
		len = ptr - from;
		strncpy( to, from, len );
		to += len;
		*to = '\0';

		from += len;

		if ( show_html ) {
			strcat( to, YAHOO_STYLE_URLON );
			to += strlen( YAHOO_STYLE_URLON );
		}

		while(( *from ) && ( ! isspace( *from ))) {
			*(to++) = *(from++);
		}
		*to = '\0';

		if ( show_html ) {
			strcat( to, YAHOO_STYLE_URLOFF );
			to += strlen( YAHOO_STYLE_URLOFF );
		}

		ptr = strstr( from, "http://" );
	}

	strcat( to, from );
	to += strlen( from );
	*to = '\0';
	strcpy( str, tmp );
}

/*
 * strip junk (like "<snd" directives out of text
 */
void strip_junk( char *str ) {
	/* fixme */
}

void tab_complete_username() {
	char *text = gtk_entry_get_text( GTK_ENTRY(chat_entry));
	char new[1025];
	char *ptr;

	if ( ! strlen( text ))
		return;

	strcpy( new, text );
	ptr = new + strlen( new ) - 1;

	while(( ptr > new ) && ( ! isspace( *ptr )))
		ptr--;

	if ( isspace( *ptr ))
		ptr++;

	if ( *ptr == '/' ) {
	} else {
		strcpy( ptr, find_user_best_match( ptr ));
	}

	gtk_entry_set_text( GTK_ENTRY(chat_entry), new );
	gtk_entry_set_position( GTK_ENTRY(chat_entry), strlen(new));
}

void capture_text_to_file( char *filename ) {
	FILE *fp;
	int  length;
	int  i;
	gchar ch;
	time_t time_now;

	fp = fopen( filename, "ab" );

	if ( ! fp )
		return;

	length = gtk_text_get_length( GTK_TEXT(chat_text) );

	time_now = time(NULL);

	fprintf( fp, "Captured at: %s", ctime(&time_now) );
	fprintf( fp, "-------------------------------------\n" );

	for( i = 0; i < length; i++ ) {
		ch = GTK_TEXT_INDEX( GTK_TEXT(chat_text), i );
		fwrite( &ch, 1, 1, fp );
	}

	fprintf( fp, "\n" );
	fclose( fp );
}

void set_setup_options() {
	GtkWidget *tmp_widget;
	GList *this_regex;
	GList *this_user;
	gchar *listitem[2];
	int row;

	/* show the correct settings */
	tmp_widget = lookup_widget( setup_window, "auto_login" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), auto_login );

	tmp_widget = lookup_widget( setup_window, "blended_colors" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_blended_colors );

	tmp_widget = lookup_widget( setup_window, "packet_debugging" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), debug_packets );

	if ( auto_away ) {
		tmp_widget = lookup_widget( setup_window, "auto_away" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "auto_away_time" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)auto_away );
	}

	tmp_widget = lookup_widget( setup_window, "custom_away_message" );
	if ( custom_away_message )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), custom_away_message );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );

	tmp_widget = lookup_widget( setup_window, "browser_command" );
	if ( browser_command )
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), browser_command );
	else
		gtk_entry_set_text(GTK_ENTRY(tmp_widget), "" );

	tmp_widget = lookup_widget( setup_window, "pm_brings_back" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), pm_brings_back );

	tmp_widget = lookup_widget( setup_window, "style_bold" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_bold );

	tmp_widget = lookup_widget( setup_window, "style_underline" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_underline );

	tmp_widget = lookup_widget( setup_window, "style_italic" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), use_italics );

	if ( buffer_size ) {
		tmp_widget = lookup_widget( setup_window, "limit_buffer_size" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "buffer_size" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)buffer_size );
	}

	if ( ping_minutes ) {
		tmp_widget = lookup_widget( setup_window, "use_ping_time" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "ping_minutes" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)ping_minutes );
	}

	if ( history_limit ) {
		tmp_widget = lookup_widget( setup_window, "use_history_limit" );
		gtk_toggle_button_set_active(
			&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), 1 );

		tmp_widget = lookup_widget( setup_window, "history_limit" );
		gtk_spin_button_set_value( GTK_SPIN_BUTTON( tmp_widget ),
			(float)history_limit );
	}

	tmp_widget = lookup_widget( setup_window, "force_lowercase" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), force_lowercase );

	tmp_widget = lookup_widget( setup_window, "show_colors" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_colors );

	tmp_widget = lookup_widget( setup_window, "show_fonts" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_fonts );

	tmp_widget = lookup_widget( setup_window, "show_html" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_html );

	tmp_widget = lookup_widget( setup_window, "show_statuses" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_statuses );

	tmp_widget = lookup_widget( setup_window, "show_enters" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), show_enters );

	tmp_widget = lookup_widget( setup_window, "profile_viewer" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), profile_viewer );

	tmp_widget = lookup_widget( setup_window, "ignore_on_mults" );
	gtk_toggle_button_set_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button), ignore_on_mults );

	tmp_widget = lookup_widget( setup_window, "fontselection1" );
	gtk_font_selection_set_font_name(
		GTK_FONT_SELECTION(tmp_widget), font_name );

	/* setup current color */
	if ( ! use_color )
		use_color = strdup( "black" );

	tmp_widget = lookup_widget( setup_window, "color_black" );
	GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active = 0;

	tmp_widget = NULL;
	if ( ! strcmp( use_color, "black" )) {
		tmp_widget = lookup_widget( setup_window, "color_black" );
	} else if ( ! strcmp( use_color, "red" )) {
		tmp_widget = lookup_widget( setup_window, "color_red" );
	} else if ( ! strcmp( use_color, "blue" )) {
		tmp_widget = lookup_widget( setup_window, "color_blue" );
	} else if ( ! strcmp( use_color, "green" )) {
		tmp_widget = lookup_widget( setup_window, "color_green" );
	} else if ( ! strcmp( use_color, "yellow" )) {
		tmp_widget = lookup_widget( setup_window, "color_yellow" );
	} else if ( ! strcmp( use_color, "purple" )) {
		tmp_widget = lookup_widget( setup_window, "color_purple" );
	} else if ( ! strcmp( use_color, "cyan" )) {
		tmp_widget = lookup_widget( setup_window, "color_cyan" );
	} else if ( ! strcmp( use_color, "olive" )) {
		tmp_widget = lookup_widget( setup_window, "color_olive" );
/* fixme, support for RGB colors
	} else if ( use_color[0] == '#' ) {
*/
	} else {
		tmp_widget = lookup_widget( setup_window, "color_black" );
	}
	GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active = 1;

	/* loop through and set lists */
	tmp_widget = lookup_widget( setup_window, "friend_list" );

	tmp_widget = lookup_widget( setup_window, "regex_list" );
	gtk_clist_set_sort_column(GTK_CLIST(tmp_widget), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(tmp_widget), 1 );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));

	this_regex = regex_list;
	while( this_regex ) {
		listitem[0] = this_regex->data;
		listitem[1] = NULL;
		row = gtk_clist_insert(GTK_CLIST(tmp_widget), 0, listitem );
		gtk_clist_set_text(GTK_CLIST(tmp_widget), row, 1, "" );

		this_regex = g_list_next( this_regex );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));

	tmp_widget = lookup_widget( setup_window, "ignore_list" );
	gtk_clist_set_sort_column(GTK_CLIST(tmp_widget), 0 );
	gtk_clist_set_auto_sort(GTK_CLIST(tmp_widget), 1 );
	gtk_clist_freeze(GTK_CLIST(tmp_widget));
	this_user = ignore_list;
	while( this_user ) {
		listitem[0] = this_user->data;
		row = gtk_clist_insert(GTK_CLIST(tmp_widget), 0, listitem );
		this_user = g_list_next( this_user );
	}
	gtk_clist_thaw(GTK_CLIST(tmp_widget));


}

void get_setup_options() {
	GtkWidget *tmp_widget;
	char *ptr;
	char tmp_filename[257];
	gchar *tmp_str;
	int i;
	FILE *fp;

	/* gather & save the settings */
	tmp_widget = lookup_widget( setup_window, "auto_login" );
	auto_login = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "blended_colors" );
	show_blended_colors = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "packet_debugging" );
	debug_packets = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "auto_away" );
	auto_away = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( auto_away ) {
		tmp_widget = lookup_widget( setup_window, "auto_away_time" );
		auto_away = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "custom_away_message" );
	if ( custom_away_message ) {
		free( custom_away_message );
		custom_away_message = NULL;
	}
	custom_away_message = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));

	tmp_widget = lookup_widget( setup_window, "browser_command" );
	if ( browser_command ) {
		free( browser_command );
		browser_command = NULL;
	}
	browser_command = strdup(gtk_entry_get_text(GTK_ENTRY(tmp_widget)));

	tmp_widget = lookup_widget( setup_window, "pm_brings_back" );
	pm_brings_back = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "style_bold" );
	use_bold = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "style_underline" );
	use_underline = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "style_italic" );
	use_italics = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "limit_buffer_size" );
	buffer_size = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( buffer_size ) {
		tmp_widget = lookup_widget( setup_window, "buffer_size" );
		buffer_size = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "use_history_limit" );
	history_limit = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( history_limit ) {
		tmp_widget = lookup_widget( setup_window, "history_limit" );
		history_limit = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "use_ping_time" );
	ping_minutes = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	if ( ping_minutes ) {
		tmp_widget = lookup_widget( setup_window, "ping_minutes" );
		ping_minutes = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON( tmp_widget ));
	}

	tmp_widget = lookup_widget( setup_window, "force_lowercase" );
	force_lowercase = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_colors" );
	show_colors = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_fonts" );
	show_fonts = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_html" );
	show_html = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_statuses" );
	show_statuses = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "show_enters" );
	show_enters = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "profile_viewer" );
	profile_viewer = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "ignore_on_mults" );
	ignore_on_mults = gtk_toggle_button_get_active(
		&(GTK_CHECK_BUTTON(tmp_widget)->toggle_button));

	tmp_widget = lookup_widget( setup_window, "fontselection1" );
	free( font_name );
	font_name = strdup( gtk_font_selection_get_font_name(
		GTK_FONT_SELECTION(tmp_widget)));
	ptr = font_name;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	free( font_family );
	font_family = strdup( ptr );
	ptr = strchr( font_family, '-' );
	*ptr = '\0';
	ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	ptr = strchr( ptr, '-' ); ptr++;
	font_size = atoi( ptr );

	/* get the user's color */
	free( use_color );
	use_color = NULL;
	if (( tmp_widget = lookup_widget( setup_window, "color_black" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "black" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_red" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "red" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_blue" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "blue" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_green" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "green" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_yellow" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "yellow" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_purple" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "purple" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_cyan" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "cyan" );
	} else if (( tmp_widget = lookup_widget( setup_window, "color_olive" )) &&
		( GTK_RADIO_BUTTON(tmp_widget)->check_button.toggle_button.active )) {
		use_color = strdup( "olive" );
	}

	if ( ! use_color )
		use_color = strdup( "black" );

	auto_away_time = time(NULL) + ( auto_away * 60 );

	/* now get and save the friends, regex, and ignore lists */
	tmp_widget = lookup_widget( setup_window, "regex_list" );
	sprintf( tmp_filename, "%s/.gyach/ignore.regex", getenv( "HOME" ));
	fp = fopen( tmp_filename, "wb" );

	if ( fp ) {
		i = 0;
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
				break;
			}

			fprintf( fp, "%s\n", tmp_str );

			i++;
		}
		fclose( fp );
	}

	tmp_widget = lookup_widget( setup_window, "ignore_list" );
	sprintf( tmp_filename, "%s/.gyach/ignore.list", getenv( "HOME" ));
	fp = fopen( tmp_filename, "wb" );

	if ( fp ) {
		i = 0;
		while(1) {
			if ( !gtk_clist_get_text(GTK_CLIST(tmp_widget), i, 0, &tmp_str )) {
				break;
			}

			fprintf( fp, "%s\n", tmp_str );

			i++;
		}
		fclose( fp );
	}

	/* now load the files we just saved */
	ignore_load();
}

/* fixme, this function is very kludgey */
void populate_room_list( void ) {
	GtkCTree *room_tree;
	GtkCTreeNode *nodes[21];
	GtkLabel *room_label;
	char file_data[65536];
	char label[129] = "";
	char *data[1];
	char *room_file;
	char *ptr;
	char *end;
	int nread;
	int len;
	int level = 1;
	int selectable;
	int fh;

	data[0] = label;
	room_tree = GTK_CTREE(lookup_widget( room_window, "room_tree" ));

	room_label =
		GTK_LABEL( lookup_widget( GTK_WIDGET(room_tree), "loading_status" ));

	gtk_label_set_text( room_label, "Retrieving Room List..." );

	while( gtk_events_pending())
		gtk_main_iteration();

	room_file = download_image( YAHOO_ROOM_LIST_URL );

	gtk_label_set_text( room_label, "Loading Room List..." );

	while( gtk_events_pending())
		gtk_main_iteration();

	fh = open( room_file, O_RDONLY );
	nread = read( fh, file_data, 65535 );
	close( fh );

	gtk_label_set_text( room_label, "Processing Room List..." );

	while( gtk_events_pending())
		gtk_main_iteration();

	file_data[nread] = '\0';

	nodes[0] = NULL;

	gtk_clist_freeze( GTK_CLIST(room_tree) );

	ptr = strchr( file_data, '<' );
	while(( ptr ) &&
		( *ptr ) &&
		( ptr < ( file_data + nread ))) {
		end = strchr( file_data, '>' );
		len = ptr - end;
		selectable = 0;
		label[0] = '\0';

		if ( ! strncmp( ptr, "<big>", 5 )) {
			/* beginning of a group, go down a level */
			ptr += 5;
			end = strchr( ptr, '<' );
			len = end - ptr;
			strncpy( label, ptr, len );
			label[len] = '\0';
			ptr = end;
		} else if ( ! strncmp( ptr, "<ul>", 4 )) {
			/* beginning of room list */
			level++;
			ptr += 4;
		} else if ( ! strncmp( ptr, "</ul>", 5 )) {
			/* beginning of room list, up a level */
			level--;
			ptr += 5;
		} else if ( ! strncmp( ptr, "<a href=\"/?room=", 16 )) {
			/* beginning of a room name */
			ptr = strstr( ptr, "\">" );
			ptr += 2;
			end = strchr( ptr, '<' );
			len = end - ptr;
			strncpy( label, ptr, len );
			label[len] = '\0';
			ptr = end;
			selectable = 1;
		}

		/* for debugging
		if ( ptr < ( file_data + nread - 31 )) {
			printf( "%5d: %c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c"
				"%c%c%c%c%c%c%c%c%c\n", ptr - file_data,
				ptr[0], ptr[1], ptr[2], ptr[3], ptr[4],
				ptr[5], ptr[6], ptr[7], ptr[8], ptr[9],
				ptr[10], ptr[11], ptr[12], ptr[13], ptr[14],
				ptr[15], ptr[16], ptr[17], ptr[18], ptr[19],
				ptr[20], ptr[21], ptr[22], ptr[23], ptr[24],
				ptr[25], ptr[26], ptr[27], ptr[28], ptr[29] );
		}

		if ( strlen( label )) {
			printf( "label: %s\n", label );
			printf( "level: %d\n", level );
		}
		fflush( stdout );
		*/

		if ( strlen( label )) {
			nodes[level] = gtk_ctree_insert_node( room_tree, nodes[level-1],
							NULL, data, 5, NULL, NULL, NULL, NULL, FALSE,
							FALSE);

			gtk_ctree_node_set_selectable( room_tree, nodes[level],
				selectable );
			gtk_ctree_node_set_row_data( room_tree, nodes[level],
				strdup( label ));
		}

		ptr = strchr( ptr + 1, '<' );
	}

	unlink( room_file );

	gtk_label_set_text( room_label, "Room List" );
	/*
	gtk_clist_column_titles_hide( GTK_CLIST(room_tree));
	*/

	gtk_ctree_set_line_style( room_tree, GTK_CTREE_LINES_NONE);
	gtk_ctree_set_expander_style( room_tree, GTK_CTREE_EXPANDER_TRIANGLE);
	gtk_clist_thaw( GTK_CLIST(room_tree) );
}

void subst_escs( char *str ) {
	char tmp[2048];
	char *from = tmp;
	char *to = str;

	if ( ! strchr( str, '\\' ))
		return;

	strcpy( tmp, str );

	while( *from ) {
		if (( *from == '\\' ) &&
			( *(from+1) == 'n' )) {
			*(to++) = '\n';
			from += 2;
		} else {
			*(to++) = *(from++);
		}
	}
	*to = '\0';
}

void history_add( char *str ) {
	static char last[129] = "";
	char *ptr;

	/* don't add the same entry twice in a row */
	if ( ! strcmp( last, str ))
		return;

	strcpy( last, str );

	history_size++;

	while ( history_size > history_limit ) {
		history_size--;
		ptr = history_list->data;
		history_list = g_list_remove( history_list, ptr );
		free( ptr );
		strcpy( history_list->data, "" );
	}

	history_current = history_size;
	history_list = g_list_append( history_list, strdup( str ));
}

char *history_next() {
	GList *this_item;

	if ( ! history_size )
		return( "" );

	history_current++;

	if ( history_current >= history_size )
		history_current = 0;

	this_item = g_list_nth( history_list, history_current );

	return( this_item->data );
}

char *history_prev() {
	GList *this_item;

	if ( ! history_size )
		return( "" );

	history_current--;

	if ( history_current < 0 )
		history_current = history_size - 1;

	this_item = g_list_nth( history_list, history_current );

	return( this_item->data );
}

void history_print() {
	GList *this_item;
	char buf[129];

	append_to_textbox( chat_window, chat_text, "*** Command History ***\n" );

	this_item = history_list;
	while( this_item ) {
		if ( strcmp( this_item->data, "" )) {
			sprintf( buf, "%s\n", (char *)this_item->data );
			append_to_textbox( chat_window, chat_text, buf );
		}
		this_item = g_list_next( this_item );
	}
	append_to_textbox( chat_window, chat_text, "*** End History ***\n" );
}
